'use strict';

var JsxSyntax = require('@babel/plugin-syntax-jsx');
var helperModuleImports = require('@babel/helper-module-imports');
var types = require('@babel/types');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var JsxSyntax__default = /*#__PURE__*/_interopDefaultLegacy(JsxSyntax);

const JSX_LIBRARY_NAME = '@whatsup/jsx';
const FRAGMENT_COMPONENT_NAME = 'Fragment';
const JSX_LIBRARY_FACTORIES = {
  HTML: 'html',
  SVG: 'svg',
  Component: 'component'
};
const SVG_TAG_REGEX = /^(?:sv|an|c(?:ir|l)|de[fs]|el|f(?:e|il|ore)|g$|ima|line|m(?:a(?:rke|sk)|etad|p)|p(?:at[ht]|oly[gl])|r[ae]|s(?:to|[wy])|t(?:ext(?:$|P)|s)|us|vie)/;
const IS_TESTING = (() => {
  try {
    return process.env.PLUGIN_TESTING === 'on';
  } catch (e) {
    return false;
  }
})();

/// <reference path="./@types/babel__helper-module-imports.d.ts" />
const VOID = types.identifier('undefined');
function index () {
  return {
    inherits: JsxSyntax__default["default"],
    visitor: {
      JSXElement(path) {
        const {
          node
        } = path;
        const {
          openingElement,
          children
        } = node;
        const {
          name,
          attributes
        } = openingElement;

        if (types.isJSXNamespacedName(name)) {
          throw new Error('Namespaced openingElement name is not supported');
        }

        if (types.isJSXMemberExpression(name)) {
          throw new Error('OpeningElement name as member expression is not supported');
        }

        replaceJSXAstToMutatorFactoryCallExpression(path, name.name, attributes, children);
      },

      JSXFragment(path) {
        const {
          node
        } = path;
        const {
          children
        } = node;
        replaceJSXAstToMutatorFactoryCallExpression(path, FRAGMENT_COMPONENT_NAME, [], children);
      }

    }
  };
}

function replaceJSXAstToMutatorFactoryCallExpression(path, name, attributes, children) {
  const factory = getFactory(name);
  const callee = createCalleeImport(path, factory);
  const type = createType(path, name);
  const uid = createUid();
  const {
    key,
    ref,
    props
  } = parseAttributes(attributes);
  const childs = parseChildren(children);
  const expression = types.callExpression(callee, [type, uid, key, ref, props, childs]);
  path.replaceWith(expression);
}

function getFactory(name) {
  const {
    HTML,
    SVG,
    Component
  } = JSX_LIBRARY_FACTORIES;
  return isComponent(name) ? Component : isSVG(name) ? SVG : HTML;
}

function createFragmentImport(path) {
  return createImport(path, JSX_LIBRARY_NAME, 'Fragment');
}

function createCalleeImport(path, factory) {
  return createImport(path, JSX_LIBRARY_NAME, factory);
}

function createImport(path, importSource, method) {
  return helperModuleImports.addNamed(path, method, importSource, {
    importedType: 'es6'
  });
}

function createType(path, name) {
  if (isFragment(name)) {
    return createFragmentImport(path);
  }

  if (isComponent(name)) {
    return types.identifier(name);
  }

  return types.stringLiteral(name);
}

function createUid() {
  return types.stringLiteral(generateUid());
}

function parseAttributes(attributes) {
  let key = VOID;
  let ref = VOID;
  const members = [];

  for (const attr of attributes) {
    if (types.isJSXSpreadAttribute(attr)) {
      const member = types.spreadElement(attr.argument);
      members.push(member);
      continue;
    }

    if (types.isJSXNamespacedName(attr.name)) {
      throw new Error('Namespaced attribute name is not supported');
    }

    const {
      name
    } = attr.name;
    const value = parseAttrValue(attr.value);

    if (value !== undefined) {
      if (name === 'key') {
        key = value;
      } else if (name === 'ref') {
        ref = value;
      } else {
        const prop = types.identifier(name);
        const member = types.objectProperty(prop, value);
        members.push(member);
      }
    }
  }

  const props = members.length ? types.objectExpression(members) : VOID;
  return {
    key,
    ref,
    props
  };
}

function parseAttrValue(value) {
  if (types.isJSXExpressionContainer(value)) {
    if (types.isJSXEmptyExpression(value.expression)) {
      // Uncoveraged. JSX attributes must only be assigned a non-empty expression
      return types.nullLiteral();
    }

    return value.expression;
  }

  if (value === null) {
    return types.booleanLiteral(true);
  }

  return value;
}

function parseChildren(children) {
  const members = [];
  const {
    length
  } = children;

  for (let i = 0; i < length; i++) {
    const child = children[i];

    if (types.isJSXText(child)) {
      let value = child.value.replace(/ +/g, ' ').replace(/ ?\n ?/g, '\n');

      if (i === 0) {
        value = value.trimStart();
      }

      if (i === length - 1) {
        value = value.trimEnd();
      }

      if (value !== '' && value !== '\n') {
        members.push(types.stringLiteral(value.replace(/\n+/g, ' ')));
      }

      continue;
    }

    if (types.isJSXElement(child) || types.isJSXFragment(child)) {
      members.push(child);
      continue;
    }

    if (types.isJSXExpressionContainer(child) && !types.isJSXEmptyExpression(child.expression)) {
      members.push(child.expression);
      continue;
    }

    if (types.isJSXSpreadChild(child)) {
      members.push(types.spreadElement(child.expression));
      continue;
    }
  }

  return members.length ? types.arrayExpression(members) : VOID;
}

function generateUid() {
  return IS_TESTING ? 'UniqueId' : (~~(Math.random() * 1e8)).toString(16);
}

function isComponent(name) {
  const firstLetter = name.charAt(0);
  return firstLetter.toUpperCase() === firstLetter;
}

function isFragment(name) {
  return name === FRAGMENT_COMPONENT_NAME;
}

function isSVG(name) {
  return SVG_TAG_REGEX.test(name);
}

module.exports = index;
//# sourceMappingURL=index.cjs.js.map
